
open DL ;
print_depth 25 ;

fun tacthm tf state =
  case Sequence.pull (tf state) of
      None => state
    | Some(x,_) => x;

fun up th = tacthm (TRYALL atac) (th RS cut_rl);
val cut2 = tacthm (etac thin_rl 2) (cut_rl RS cut_rl);
val revcut2 = tacthm (etac thin_rl 2) (revcut_rl RSN (2, revcut_rl));
(*
val tacthm = fn : tactic -> thm -> thm
val up = fn : thm -> thm - unify premises
val cut2 = "[| [| PROP ?psi; PROP ?psi1 |] ==> PROP ?theta;
    PROP ?psi1; PROP ?psi |] ==> PROP ?theta" : thm
val revcut2 =
  "[| PROP ?V; PROP ?V1; [| PROP ?V; PROP ?V1 |] ==> PROP ?W |] ==> PROP ?W"
    : thm
*)
val meta_iffI = prove_goal DL.thy 
    "[| PROP P ==> PROP Q; PROP Q ==> PROP P |] ==> PROP P == PROP Q"
  (fn [prem1,prem2] => [rtac (equal_intr prem1 prem2) 1]);

(* mii : thm -> thm -> thm, takes P ==> Q and Q ==> P to P == Q,
  like equal_intr, but also does necessary resolution *)
fun mii ded1 ded2 = tacthm (atac 1 THEN atac 1) ([ded1, ded2] MRS meta_iffI) ;

val meta_iffD2 = prove_goal DL.thy "[| PROP P==PROP Q; PROP Q |] ==> PROP P"
  (fn [prem1,prem2] => [rewtac prem1, rtac prem2 1]);

val meta_iffD1 = symmetric_thm RS meta_iffD2;

fun md1 th = th RS meta_iffD1;
fun md2 th = th RS meta_iffD2;

val trans3 = transitive_thm RS transitive_thm ;
val trans4 = transitive_thm RS trans3 ;

(* reverses sequents: easier than editing! *)
fun seqrev ((tt as Const ("op |-", _)) $ L $ R) = tt $ R $ L 
  | seqrev ((mo as Const (_, _)) $ L $ R) = mo $ seqrev L $ seqrev R 
  | seqrev ((mo as Const (_, _)) $ R) = mo $ seqrev R ;
(* print a term *)
fun pt thy = Sign.string_of_term (sign_of thy) ;
fun prt thy = pt thy o seqrev o #prop o rep_thm ;

(* bind_st_thm : string * string * thm -> unit
  like bind_thm, but also gives statement of theorem (ie you can choose
  the variable names; won't work if premises are themselves meta-level
  implications); also will prove, in simple cases, a weaker theorem *)
fun bind_st_thm (name, stmt, thm) = 
  qed_goal name (theory_of_thm thm) stmt 
    (fn prems => [rtac (prems MRS thm) 1]) ;

fun tn [e] = e
  |  tn (e::el) = [e, tn el] MRS transitive_thm ;

(* Derived display equivalences *)
bind_thm ("ssAS", [sA, sS] MRS transitive_thm) ;
bind_thm ("ssS", [ssA, ssAS] MRS transitive_thm) ;
bind_thm ("ca1", [ca2, ssA, symmetric ca2, symmetric cs2] MRS trans4) ;
bind_thm ("cs1", [cs2, ssS, symmetric cs2, symmetric ca2] MRS trans4) ;
bind_thm ("sbs", [sS, symmetric blob, sA] MRS trans3) ;
bind_thm ("dcp", [ssA, sA] MRS transitive_thm) ;
bind_st_thm ("cbdista", "(* ($X, $Y)) |- $Z == * $Y, * $X |- $Z", (* () *)
  [sA, cs1, symmetric cs2, cs1] MRS trans4) ;
bind_st_thm ("cbdists", "$X |- * ($Y, $Z) == $X |- * $Z, * $Y",
  [sS, ca1, symmetric ca2, ca1] MRS trans4) ;

(* to derive ca2, cs2, from ca1, cs1
bind_thm ("ca2", [ca1, ssA, symmetric ca1, symmetric cs1] MRS trans4) ;
bind_thm ("cs2", [cs1, ssS, symmetric cs1, symmetric ca1] MRS trans4) ;
*)

(* Derived classical modal structural rules *)
bind_thm ("pAe", mii pA pA) ;
bind_thm ("pS", md1 cs1 RS pA RS md2 cs2) ;
bind_thm ("pSe", mii pS pS) ;
(* bind_thm ("pSe", tn [dcp, cbdista, pAe, symmetric cbdista, symmetric dcp]) ; *)
bind_thm ("mra", mla RS pA) ;
bind_thm ("mrs", md1 dcp RS mla RS md2 cbdista RS md2 dcp) ;
bind_thm ("mls", mrs RS pS) ;
bind_thm ("cS", md1 dcp RS md1 cbdista RS cA RS md2 dcp) ;
bind_st_thm ("aS", "$Z |- $W, ($X, $Y) == $Z |- ($W, $X), $Y",
  tn [dcp, cbdista, symmetric cs2, cbdista, cs2, 
    symmetric ( tn [cs1, ca1, cbdista, cs1, aA])]) ; 
bind_thm ("ira", pA RS ila) ;
bind_thm ("irs", pS RS ils) ;
bind_st_thm ("exa", "$X |- $I ==> $X |- $Z", mrs RS ils) ;
bind_st_thm ("exs", "$I |- $X ==> $Z |- $X", mla RS ira) ;
bind_st_thm ("sis", "$X |- *$I ==> $X |- $I", md1 sS RS exs RS md2 dcp) ;
bind_st_thm ("sia", "(*$I) |- $X ==> $I |- $X", md1 sA RS exa RS md2 dcp) ;
    (* () *)
bind_st_thm ("nA", "$I |- $X ==> @$I |- $X", exs) ;
bind_st_thm ("n2a", "$I |- $X ==> *@* $I |- $X", exs) ;
bind_st_thm ("nS", "$X |- $I ==> $X |- *@* $I", exa) ;
bind_st_thm ("n2s", "$X |- $I ==> $X |- @$I", exa) ;
(* ex and n rules, right from left
bind_thm ("exs", md1 dcp RS sis RS exa RS md2 dcp) ;
bind_st_thm ("nS", "$X |- $I ==> $X |- *@* $I",
  md1 dcp RS sia RS nA RS md2 blob RS exs RS md1 blob RS md1 sS) ;
bind_thm ("n2s",
    md1 dcp RS sia RS n2a RS md2 dcp RS md1 blob RS sis RS md2 blob) ;
*)

use "DL.tac" ;

qed_goal "bldists" DL.thy "$X |- @$Y, @$Z ==> $X |- @($Y, $Z)"
  (fn [prem] => [
    rtac cS 1,
    d_b_tac "-l@" (rtac mrs) 1,
    d_b_tac "-r@" (rtac mls) 1,
    rtac prem 1]) ;

qed_goal "bldista" DL.thy "(@$X, @$Y) |- $Z ==> @($X, $Y) |- $Z" (* () *)
  (fn [prem] => [
    rtac cA 1,
    d_b_tac "|l@" (rtac mra) 1,
    d_b_tac "|r@" (rtac mla) 1,
    rtac prem 1]) ;

qed_goal "sbsdists" DL.thy "$X |- *@* $Y, *@* $Z ==> $X |- *@* ($Y, $Z)"
  (fn [prem] => [
    rtac cS 1,
    d_b_tac "-l*@*" (rtac mrs) 1,
    d_b_tac "-r*@*" (rtac mls) 1,
    rtac prem 1]) ;

qed_goal "sbsdista" DL.thy "(*@* $X, *@* $Y |- $Z) ==> *@*($X, $Y) |- $Z" (*()*)
  (fn [prem] => [
    rtac cA 1,
    d_b_tac "|l*@*" (rtac mra) 1,
    d_b_tac "|r*@*" (rtac mla) 1,
    rtac prem 1]) ;

use "wk.tac" ;
