(* DL.tac *)

val structT = Type ("structure", []) ;
val struct_unT = Type ("fun", [structT, structT]) ;
val struct_binT = Type ("fun", [structT, struct_unT]) ;

(* find symmetry rule to display part of tm, asp indicated by str
  asp is ~1 means looking at whole sequent
  asp 0 or 1 means tm is on left or right hand side of turnstile *)

fun draux [] tm _ = reflexive_thm
  (* applies dcp to delete * on both sides *)
  | draux (t::"*"::str) 
      (Const ("Sequent", _) $ (Const ("*", _) $ L) $ (Const ("*", _) $ R)) ~1
    = (case t of "|" => tn [symmetric dcp, draux str L 1]
	      | "-" => tn [symmetric dcp, draux str R 0])
  | draux ("|"::str) (Const ("Sequent", _) $ L $ R) ~1 = draux str L 0
  | draux ("-"::str) (Const ("Sequent", _) $ L $ R) ~1 = draux str R 1
  (* deletes ** without adding it to other side *)
  | draux ("*"::"*"::str) (Const ("*", _) $ (Const ("*", _) $ U)) asp 
    = tn [symmetric (nth ([ssA, ssS], asp)), draux str U asp]
  | draux ("*"::str) (Const ("*", _) $ U) asp
    = tn [nth ([sA, sS], asp), draux str U (1-asp)]
  (* applies dcp to add * on both sides *)
  | draux ("x"::str) U asp (* to display *A *)
    = tn [dcp, draux str (Const ("*", struct_unT) $ U) (1-asp)]
  | draux ("@"::str) (Const ("@", _) $ U) asp
    = tn [nth ([symmetric blob, blob], asp), draux str U asp]
  | draux ("l"::str) (Const (",", _) $ L $ (Const ("*", _) $ U)) asp 
    = tn [symmetric (nth ([cs2, ca1], asp)), draux str L asp]
  | draux ("r"::str) (Const (",", _) $ (Const ("*", _) $ U) $ R) asp 
    = tn [symmetric (nth ([cs1, ca2], asp)), draux str R asp]
  | draux ("l"::str) (Const (",", _) $ L $ R) asp
    = tn [nth ([ca1, cs2], asp), draux str L asp]
  | draux ("r"::str) (Const (",", _) $ L $ R) asp
    = tn [nth ([ca2, cs1], asp), draux str R asp] ;

fun disp_rule str tm = draux (explode str) tm ~1 ;

(* disp_tac = fn : string -> thm -> thm Sequence.seq
  for backwards proof, displays part of mth subgoal identified by str *)
fun disp_tac str m state =
    rtac (md2 (disp_rule str (nth (prems_of state, m-1)))) m state ;
fun bdt str = by (disp_tac str 1) ;


(* THENR : int -> int -> (int -> tactic) -> tactic
  like ALLGOALS but with a limited range of subgoals,
  does tf j for j = n downto m *)
fun THENR m n tf = 
  if n = m then tf m 
    else if n < m then all_tac
    else tf n THEN THENR m (n-1) tf ;

infix 1 THENEXP ; 

(* THENEXP : (int -> tactic) * (int -> tactic) -> int -> tactic
  composition of tactic functions tacf1,tacf2, where 
  tacf1,tacf2 : int -> tactic,
  and the integer argument is a subgoal number;
  if tacf1 m replaces subgoal m by a number of subgoals,
  tacf2 is then applied to each of those subgoals *)
fun (tacf1 THENEXP tacf2) sg state = 
  let fun tac2 newstate = 
	(THENR sg (nprems_of newstate - nprems_of state + sg) tacf2) newstate
  in (tacf1 sg THEN tac2) state end ;

(* d_b_tac = fn : string -> (int -> tactic) -> int -> thm -> thm Sequence.seq
  displays selected part, then acts on resulting state  with tactic tf m,
  (eg tf is (rtac rule)) then reverses display operation on all resulting
  subgoals *)
fun d_b_tac str tf sg state = 
  let val dr = disp_rule str (nth (prems_of state, sg-1))
  in (rtac (md2 dr) THEN' tf THENEXP rtac (md1 dr)) sg state end ;
fun bdbr str thm = by (d_b_tac str (rtac thm) 1) ;

(* glob_down, glob_at, new_glob_at :
  term -> int -> (term -> int -> (int -> tactic) list) -> int -> tactic
  used for glob_tac;
  glob_down tm asp looks inside subterm tm, displaying subterms 
  glob_at tm asp looks at subterm tm
  new_glob_at _ asp looks at subterm tm, assuming it has been changed,
  recalculates it, and repeats glob_at *)
fun glob_down (Const (",", _) $ L $ R) asp act 
    = rtac (md2 (nth ([ca1, cs2], asp))) 
      THENEXP glob_at L asp act
      THENEXP rtac (md1 (nth ([ca1, cs2], asp)))
      THENEXP rtac (md2 (nth ([ca2, cs1], asp)))
      THENEXP glob_at R asp act
      THENEXP rtac (md1 (nth ([ca2, cs1], asp)))
  | glob_down (Const ("*", _) $ U) asp act 
    = rtac (md2 (nth ([sA, sS], asp)))
      THENEXP glob_at U (1-asp) act
      THENEXP rtac (md1 (nth ([sA, sS], asp)))
  | glob_down (Const ("@", _) $ U) asp act 
    = rtac ((nth ([md1, md2], asp)) blob)
      THENEXP glob_at U asp act
      THENEXP rtac ((nth ([md2, md1], asp)) blob)
  | glob_down tm asp act = (fn i => all_tac)
and glob_at tm asp act = 
    case act tm asp of [] => glob_down tm asp act 
      | [tacf] => tacf THENEXP new_glob_at tm asp act
and new_glob_at tm asp act sg state = 
    let val (Const ("Sequent", _) $ L $ R) =
	nth (prems_of state, sg-1)
    in glob_at (nth ([L,R], asp)) asp act sg state end ;

(* glob_tac : (term -> int -> (int -> tactic) list) -> int -> tactic
  act tm asp is either [] or [tacf], where 
    tm is a subterm, and asp indicates antecedent or succedent part
    tacf sg is a tactic (for sg a subgoal number)
    tacf should be a tactic which changes only tm when it is displayed as the
    antecedent or succedent (as indicated by asp)
  glob_tac act sg displays every subterm of the subgoal, and
  applies the tactic, if any, given by act tm asp, at every subterm *)
fun glob_tac act sg state =
    let val (Const ("Sequent", _) $ L $ R) =
      nth (prems_of state, sg-1)
    in (glob_at L 0 act THENEXP glob_at R 1 act) sg state end ;

(* tfprlors : (('a -> bool) * thm) list -> 'a -> (int -> tactic) list
  (used with 'a = term)
  in tfprlors patrules tm
  patrules is a list of (pattern) tests and theorems
  if tm doesn't satisfy any test, tfprlors returns []
  otherwise, take the first test satisfied by tm, let thm be the
  corsesponding theorem, the tfprlors returns [rtac thm] *)
fun tfprlors [] _ = []
  | tfprlors ((patfn, rule) :: patrules) tm =
    if patfn tm then [rtac rule]
    else tfprlors patrules tm ;

(* tfpr : (('a -> bool) * thm) list list -> 'a -> int -> (int -> tactic) list
  (used with 'a = term)
  in tfpr [apr, spr] tm asp, selects apr or spr according as asp indicates
  antecedent or succedent part, then applies tfprlors above *)
fun tfpr aspr tm asp = tfprlors (nth (aspr, asp)) tm ;

fun bgt patrules = by (ALLGOALS (glob_tac (tfpr patrules))) ;

fun pat_ss (Const ("*", _) $ (Const ("*", _) $ U)) = true | pat_ss _ = false ;

fun pat_Il (Const (",", _) $ Const ("I", _) $ R) = true | pat_Il _ = false ;

fun pat_Ir (Const (",", _) $ L $ Const ("I", _)) = true | pat_Ir _ = false ;

fun pat_sc (Const ("*", _) $ (Const (",", _) $ L $ R)) = true
  | pat_sc _ = false ;

val pr_id = [[ (pat_Il, mla), (pat_Ir, mra) ],
	    [ (pat_Il, mls), (pat_Ir, mrs) ] ] ;
	  
val pr_sc = [[ (pat_sc, md2 cbdista) ], [ (pat_sc, md2 cbdists) ] ] ;

val pr_ss = [[ (pat_ss, md1 ssA) ], [ (pat_ss, md1 ssS) ] ] ;

fun pr_append ([apr1, spr1], [apr2, spr2]) = [apr1 @ apr2, spr1 @ spr2] ;
val pr_appl = foldr1 pr_append ;
