(* dKt.tac *)

fun pat_fsunop opstr (Const ("Structform", _) $ (Const (str, _) $ U)) = 
    str = opstr 
  | pat_fsunop _ _ = false ;
fun pat_fsbinop opstr (Const ("Structform", _) $ (Const (str, _) $ L $ R)) = 
    str = opstr 
  | pat_fsbinop _ _ = false ;

val pr_intr = [[
  (pat_fsbinop "op &", anda),
  (pat_fsbinop "op v", orA),
  (pat_fsbinop "op ->", impA),
  (pat_fsunop "~", nota)
  ], [
  (pat_fsbinop "op &", andS),
  (pat_fsbinop "op v", ors),
  (pat_fsbinop "op ->", impS),
  (pat_fsunop "~", nots)
  ]] ;

val pr_mod = [[
  (pat_fsunop "wdm", wdma),
  (pat_fsunop "bdm", bdmA)
  ], [
  (pat_fsunop "wbx", wbxS),
  (pat_fsunop "bbx", bbxs)
  ]] ;

(* test
goal thy "*@( **A, **B) |- *@**C, **@ D" ;
goal thy "*I, @( **A, I, **B) |- *@(I, **C), I, *( *@ D, I)" ;
goal thy "~(F -> A) & ~B |- ~C & D -> ~E v G" ;
goal thy "~(~C & D) & ( ~E v G) |- ~(F -> A) v ~B" ;
by (glob_tac (tfpr pr_intr) 1) ;
by (ALLGOALS (glob_tac (tfpr pr_ss))) ;
goal thy "~C & D -> ~E v G |- ~(F -> A) & ~B" ;
*)

(* flaux : term -> string list -> int -> (term * string * int) list
    -> (term * string * int) list
  auxiliary function for form_list, 
  flaux tm loc asp sofar looks at subterm tm, whose location in the
  sequent is indicated by loc, with asp indicating whether it is
  antecedent or succedent, and sofar is the list of formulae already found *)
fun flaux (Const (",", _) $ L $ R) loc asp sofar =
      flaux R ("r"::loc) asp (flaux L ("l"::loc) asp sofar)
  | flaux (Const ("*", _) $ U) loc asp sofar =
      flaux U ("*"::loc) (1-asp) sofar
  | flaux (Const ("Structform", _) $ F) loc asp sofar =
      (F, implode (rev loc), asp)::sofar
  | flaux tm _ _ sofar = sofar ;

(* form_list : term -> (term * string * int) list
  creates list of formulae in a sequent, list is of triples
  (term, string, asp), where term is the formula, without initial
  Structform (_), string indicates the location in the sequent,
  and asp = 0 (1) for antecedent (succedent) part 
  Note - doesn't look underneath @, because it is for doing simpae proof
  by weakening and identity of formulae *)
fun form_list (Const ("Sequent", _) $ L $ R) =
    flaux R ["-"] 1 (flaux L ["|"] 0 []) ;

fun tm_cmp (O1 $ U1) (O2 $ U2) = tm_cmp O1 U1 andalso tm_cmp U1 U2
  | tm_cmp op1 op2 = (op1 = op2) ;

fun tsi_cmp (tm1, str1, int1) (tm2, str2, int2) = tm_cmp tm1 tm2 

fun tsi_match tsi (tsil::tsis) = 
    (if tsi_cmp tsi tsil then [tsi, tsil] else tsi_match tsi tsis)
  | tsi_match tsi [] = [] ;

fun first_match (tsi::tsis) = 
    let val tsim = tsi_match tsi tsis 
    in if tsim = [] then first_match tsis else tsim end 
  | first_match [] = [] ; 

fun weak_tac sg state = 
  let val fm = first_match (form_list (nth (prems_of state, sg-1)))
  in case fm of [] => Sequence.null
    | [(tm1, str1, int1), (tm2, str2, int2)] =>
      (keep_tac [str1, str2]
	THEN' glob_tac (tfpr (pr_append (pr_ss, pr_sc)))
	THEN' resolve_tac [idf, idf RS md1 dcp,
	  idf RS mla RS md1 ca1, idf RS mra RS md1 ca2,
	  idf RS mrs RS md1 cs1, idf RS mls RS md1 cs2])
	sg state 
  end ;

(* test
goal thy "P -> (Q -> R) |- P & Q -> R";
bgt pr_intr ;
by (ALLGOALS weak_tac) ;
goal thy "P -> (Q -> R) |- ~( ~P & P)";
bgt pr_intr ;
by (ALLGOALS weak_tac) ;
*)
