
(* redstrs : ''a list list -> ''a -> ''a list list
  tests a list of lists against a value tch;
  where tch is the first element in a list it is deleted;
  otherwise that list is deleted *)
fun redstrs ((ch::str)::strs) tch = 
    if ch = tch then str::(redstrs strs tch)
    else redstrs strs tch 
  | redstrs [] tch = [] ;

(* ktes : string list -> term -> int -> int -> tactic
  in ktes loc tm asp, replaces subterm tm by I
    loc is string, exploded and reversed, which indicates location of tm
      in the sequent
    asp is ~1, 0 or 1 as tm is the sequent, or an antecedent or succedent part
    *)
fun ktes [] (Const ("Sequent", _) $ L $ R) ~1 =
    rtac (exa RS exs)
  | ktes loc _ asp = d_b_tac (implode (rev loc)) (rtac (nth ([exs, exa], asp)));

(* ktaux : string list list -> string list -> term -> int -> int -> tactic
  in ktaux strs loc tm asp,
    strs is list of strings indicating locations of subterms of tm which
      are to be kept (ie not replaced by I)
    tm is current subterm, loc and asp as for ktes above *)
fun ktaux [[]] loc tm asp = (fn i => all_tac)
  | ktaux [] loc tm asp = ktes loc tm asp
  | ktaux strs loc tm asp =
    let
      fun doch lch tm asp = 
	let val rstrs = redstrs strs lch
	in ktaux rstrs (lch::loc) tm asp end 
    in 
      case tm of
	(Const ("Sequent", _) $ L $ R) =>
	  doch "|" L 0 THEN' doch "-" R 1
       | (Const (",", _) $ L $ R) => doch "l" L asp THEN' doch "r" R asp
       | (Const ("*", _) $ U) => doch "*" U (1-asp) 
       | (Const ("@", _) $ U) => doch "@" U asp
    end ;
	
(* keep_tac : string list -> int -> tactic
  replaces all but indicated subterms by I, then eliminates I where possible;
  in keep_tac strs, strs indicates which subterms are to be kept *)
fun keep_tac strs sg state = ((ktaux ((map explode) strs) [] 
  (nth (prems_of state, sg-1)) ~1) THEN' glob_tac (tfpr pr_id)) sg state ;

(* test
goal thy "*@( **A, **B) |- *@**C, **@ D" ;
by (keep_tac ["|*@l*", "-r*"] 1) ;
goal thy "*@( **A, **B) |- *@**C, **@ D" ;
by (keep_tac ["|*@l*", "-l*"] 1) ;
goal thy "*@( **A, **B) |- *@**C, **@ D" ;
by (keep_tac ["|*@r*", "-r*"] 1) ;
goal thy "*@( **A, **B) |- *@**C, **@ D" ;
by (keep_tac ["|*@r*", "-l*"] 1) ;
*)
